# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
import inspect
from typing import (
    Any,
    Callable,
    Dict,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import jax
import jax.numpy as jnp
import jax.random as jr
import jax.tree_util as jtu
import numpy as np

import diffrax
import optax
from flax import nnx

__all__ = [
    "flow_matching_step",
    "interpolate_samples",
    "evaluate_velocity_field",
    "curvature",
    "gaussian_nll",
]

DivState = Tuple[jax.Array, jax.Array]  # velocity, divergence
Batch = Dict[Literal["t", "x_t", "v_t", "cond"], jax.Array]


def flow_matching_step(
    model: nnx.Module,
    optimizer: nnx.Optimizer,
    batch: Batch,
    *,
    loss_fn: Callable[[jax.Array, jax.Array], jax.Array] = optax.squared_error,
    model_callback_fn: Optional[Callable[[nnx.Module], None]] = None,
    rngs: Optional[nnx.Rngs] = None,
) -> Dict[Literal["loss", "grad_norm"], jax.Array]:
  """Perform a flow matching step.

  Args:
    model: Velocity field with a signature ``(t, x_t, cond, rngs=...) -> v_t``.
    optimizer: Optimizer.
    batch: Batch containing the following elements:

      - ``'t'`` - time, array of shape ``[batch,]``.
      - ``'x_t'`` - position, array of shape ``[batch, ...]``.
      - ``'v_t'`` - target velocity, array of shape ``[batch, ...]``.
      - ``'cond'`` - condition (optional), array of shape ``[batch, ...]``.
    loss_fn: Loss function with a signature ``(pred, target) -> loss``.
    model_callback_fn: Function with a signature ``(model) -> None``, e.g., to
      update an :class:`~ott.neural.networks.velocity_field.EMA` of the model.
    rngs: Random number generator used for, e.g., dropout, passed to the model.

  Returns:
    Updates the parameters in-place and returns the loss and the gradient norm.
  """

  def compute_loss(model: nnx.Module, rngs: nnx.Rngs) -> jax.Array:
    t, x_t, v_t = batch["t"], batch["x_t"], batch["v_t"]
    cond = batch.get("cond")
    v_pred = model(t, x_t, cond, rngs=rngs)
    return loss_fn(v_pred, v_t).mean()

  loss, grads = nnx.value_and_grad(compute_loss)(model, rngs)
  if "model" in inspect.signature(optimizer.update).parameters:
    optimizer.update(model, grads)
  else:
    # for flax version < 0.11.0
    optimizer.update(grads)
  grad_norm = optax.global_norm(grads)

  if model_callback_fn is not None:
    model_callback_fn(model)

  return {"loss": loss, "grad_norm": grad_norm}


def interpolate_samples(
    rng: jax.Array,
    x0: jax.Array,
    x1: jax.Array,
    cond: Optional[jax.Array] = None,
    *,
    time_sampler: Optional[Callable[[jax.Array, Tuple[int], jnp.dtype],
                                    jax.Array]] = None
) -> Batch:
  """Sample time and interpolate.

  Args:
    rng: Random number generator.
    x0: Source samples at :math:`t_0`, array of shape ``[batch, ...]``.
    x1: Target samples at :math:`t_1`, array of shape ``[batch, ...]``.
    cond: Condition.
    time_sampler: Time sampler with signature ``(rng, shape, dtype) -> time``.

  Returns:
    Dictionary containing the following values:

    - ``'t'`` - time, array of shape ``[batch,]``.
    - ``'x_t'`` - position :math:`x_t`, array of shape ``[batch, ...]``.
    - ``'v_t'`` - target velocity :math:`x_1 - x_0`,
      array of shape ``[batch, ...]``.
    - ``'cond'`` - condition (optional), array of shape ``[batch, ...]``.
  """
  if time_sampler is None:
    time_sampler = jr.uniform

  batch_size = len(x0)
  t = time_sampler(rng, (batch_size,), x0.dtype)
  assert t.shape == (batch_size,), (t.shape, (batch_size,))
  t_ = jnp.expand_dims(t, axis=range(1, x0.ndim))

  batch = {
      "t": t,
      "x_t": (1.0 - t_) * x0 + t_ * x1,
      "v_t": x1 - x0,
  }
  if cond is not None:
    batch["cond"] = cond
  return batch


def evaluate_velocity_field(
    model: nnx.Module,
    x: Union[jax.Array, Any],
    cond: Optional[jax.Array] = None,
    *,
    t0: float = 0.0,
    t1: float = 1.0,
    reverse: bool = False,
    num_steps: Optional[int] = None,
    solver: Optional[diffrax.AbstractSolver] = None,
    save_trajectory_kwargs: Optional[Dict[str, Any]] = None,
    save_velocity_kwargs: Optional[Dict[str, Any]] = None,
    **kwargs: Any,
) -> diffrax.Solution:
  """Solve an ODE.

  Args:
    model: Velocity field with a signature ``(t, x_t, cond) -> v_t``.
    x: Initial point of shape ``[*dims]``.
    cond: Condition of shape ``[*cond_dims]``.
    t0: Start time of the integration.
    t1: End time of the integration.
    reverse: Whether to integrate from :math:`t_1` to :math:`t_0`.
    num_steps: Number of steps used for solvers with a constant step size.
    solver: ODE solver. If :obj:`None` and ``step_size = None``,
      use :class:`~diffrax.Dopri5`. Otherwise use :class:`~diffrax.Euler`.
    save_velocity_kwargs: Keyword arguments for :class:`~diffrax.SubSaveAt`
      used to store the velocities along the integration path.
      The velocity will be saved in :class:`out.ys['v_t'] <diffrax.Solution>`.
    save_trajectory_kwargs: Keyword arguments for :class:`~diffrax.SubSaveAt`
      used to store the positions along the integration path.
      The trajectory will be saved in :class:`out.ys['x_t'] <diffrax.Solution>`.
    kwargs: Keyword arguments for :func:`~diffrax.diffeqsolve`.

  Returns:
    The ODE solution.
  """
  if isinstance(num_steps, int):
    step_size = 1.0 / num_steps
    stepsize_controller = diffrax.ConstantStepSize()
    solver = diffrax.Euler() if solver is None else solver
    kwargs["max_steps"] = num_steps
  else:
    step_size = None
    stepsize_controller = diffrax.PIDController(rtol=1e-5, atol=1e-5)
    solver = diffrax.Dopri5() if solver is None else solver

  if reverse:
    step_size = None if step_size is None else -step_size
    t0, t1 = t1, t0

  default_velocity_fn = jtu.Partial(_velocity, model=model)
  # internally, we allow for passing custom velocity functions:
  # this is used when computing the gaussian NLL, as we need to
  # both integrate the state and the divergence of the velocity field
  velocity_fn = kwargs.pop("_velocity_fn", default_velocity_fn)

  subs = {}
  if save_velocity_kwargs:
    saveat = diffrax.SubSaveAt(fn=default_velocity_fn, **save_velocity_kwargs)
    subs["v_t"] = saveat
  if save_trajectory_kwargs:
    saveat = diffrax.SubSaveAt(
        fn=lambda _, x_t, __: x_t, **save_trajectory_kwargs
    )
    subs["x_t"] = saveat

  if subs:
    kwargs["saveat"] = diffrax.SaveAt(subs=subs)

  return diffrax.diffeqsolve(
      diffrax.ODETerm(velocity_fn),
      t0=t0,
      t1=t1,
      y0=x,
      args=cond,
      solver=solver,
      dt0=step_size,
      stepsize_controller=stepsize_controller,
      **kwargs,
  )


def curvature(
    model: nnx.Module,
    x0: jax.Array,
    cond: Optional[jax.Array] = None,
    *,
    ts: Union[int, jax.Array, Sequence[float]],
    drop_last_velocity: Optional[bool] = None,
    loss_fn: Callable[[jax.Array, jax.Array], jax.Array] = optax.squared_error,
    **kwargs: Any,
) -> Tuple[jax.Array, diffrax.Solution]:
  """Compute the curvature :cite:`lee:23`.

  Also known as straightness in :cite:`liu:22`.

  Args:
    model: Velocity field with a signature ``(t, x_t, cond) -> v_t``.
    x0: Initial point of shape ``[*dims]``.
    cond: Condition of shape ``[*cond_dims]``.
    ts: Time points at which velocities are computed and stored.
      If :class:`int`, use linearly-spaced interval ``[t0, t1]``
      with ``ts`` steps.
    drop_last_velocity: Whether to remove the velocity at ``ts[-1]``.
      when computing the curvature. If :obj:`None`, don't include it when
      ``ts[-1] == 1.0``.
    loss_fn: Loss function with a signature ``(pred, target) -> loss``.
    kwargs: Keyword arguments for :func:`evaluate_velocity_field`.

  Returns:
    The curvature and the ODE solution.
  """
  if isinstance(ts, int):
    assert ts > 0, f"Number of steps must be positive, got {ts}."
    t0, t1 = kwargs.get("t0", 0.0), kwargs.get("t1", 1.0)
    ts = np.linspace(t0, t1, ts)
  if drop_last_velocity is None:
    drop_last_velocity = ts[-1] == 1.0

  sol = evaluate_velocity_field(
      model,
      x0,
      cond,
      reverse=False,
      save_trajectory_kwargs={"t1": True},  # save only at `t1`
      save_velocity_kwargs={"ts": ts},  # save `v_t` at specified times
      **kwargs,
  )
  x1 = sol.ys["x_t"][-1]
  v_t = sol.ys["v_t"][:-1] if drop_last_velocity else sol.ys["v_t"]

  steps = len(ts) - drop_last_velocity
  assert x0.shape == x1.shape, (x0.shape, x1.shape)
  assert v_t.shape == (steps, *x0.shape), (v_t.shape, (steps, *x0.shape))

  ref_velocity = (x1 - x0)
  curv = jax.vmap(loss_fn, in_axes=[0, None])(v_t, ref_velocity).mean()
  return curv, sol


def gaussian_nll(
    model: nnx.Module,
    x1: jax.Array,
    cond: Optional[jax.Array] = None,
    *,
    noise: Optional[jax.Array] = None,
    stddev: float = 1.0,
    **kwargs: Any,
) -> Tuple[jax.Array, diffrax.Solution]:
  """Compute the Gaussian negative log-likelihood.

  Args:
    model: Velocity model with a signature ``(t, x_t, cond) -> v_t``.
    x1: Initial point of shape ``[*dims]``.
    cond: Condition ``[*cond_dims]``.
    noise: Array of shape ``[num_noise_samples, ...]`` used for the Hutchinson's
      trace estimate of the divergence of the velocity field. If :obj:`None`,
      compute the exact divergence using :func:`jax.jacrev`.
    stddev: Standard deviation of the Gaussian distribution.
    kwargs: Keyword arguments for :func:`evaluate_velocity_field`.

  Returns:
    The Gaussian negative log-likelihood in bits-per-dimension.
  """
  if noise is not None:
    _, *noise_shape = noise.shape  # [batch, ...]
    assert x1.shape == tuple(noise_shape), (x1.shape, noise_shape)
    velocity_fn = functools.partial(_hutchinson_divergence, h=noise)
  else:
    velocity_fn = _exact_divergence

  sol = evaluate_velocity_field(
      model,
      (x1, jnp.zeros([])),  # initial point, divergence
      cond,
      reverse=True,
      saveat=diffrax.SaveAt(t1=True),
      save_trajectory_kwargs=None,
      save_velocity_kwargs=None,
      _velocity_fn=jtu.Partial(velocity_fn, model=model),
      **kwargs,
  )

  x0, neg_int01_div_v = sol.ys
  assert x0.shape == (1, *x1.shape), (x0.shape, (1, *x1.shape))
  assert neg_int01_div_v.shape == (1,), neg_int01_div_v.shape

  k = np.prod(x0.shape)
  logp0_x0 = -0.5 * ((x0 / stddev) ** 2).sum()
  logp0_x0 = logp0_x0 - 0.5 * k * jnp.log(2.0 * jnp.pi) - k * jnp.log(stddev)
  nll = -(logp0_x0 + neg_int01_div_v[0])
  return nll, sol


def _velocity(
    t: jax.Array, x_t: jax.Array, cond: Optional[jax.Array], model: nnx.Module
) -> jax.Array:
  cond = None if cond is None else cond[None]
  return model(t[None], x_t[None], cond).squeeze(0)


def _exact_divergence(
    t: jax.Array, state_t: DivState, cond: Optional[jax.Array], *,
    model: nnx.Module
) -> DivState:

  def divergence_v(
      t: jax.Array, x: jax.Array, cond: Optional[jax.Array]
  ) -> jax.Array:
    # divergence of fwd velocity field
    jacobian = jax.jacrev(_velocity, argnums=1)(t, x, cond, model)
    jacobian = jacobian.reshape(np.prod(x.shape), np.prod(x.shape))
    return jnp.trace(jacobian)

  x_t, _ = state_t
  v_t = _velocity(t, x_t, cond, model=model)
  div_t = divergence_v(t, x_t, cond)
  return v_t, div_t


def _hutchinson_divergence(
    t: jax.Array, state_t: DivState, cond: Optional[jax.Array], *,
    model: nnx.Module, h: jax.Array
) -> DivState:
  x_t, _ = state_t
  v_t, vjp = jax.vjp(lambda x: _velocity(t, x, cond, model=model), x_t)
  (Dvh,) = jax.vmap(vjp)(h)
  div_t = jax.vmap(jnp.vdot, in_axes=[0, 0])(h, Dvh).mean()
  return v_t, div_t



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Modified from: https://github.com/openai/guided-diffusion/blob/22e0df8183507e13a7813f8d38d51b072ca1e67c/guided_diffusion/unet.py."""
import abc
import functools
import math
from typing import Any, Literal, Optional, Tuple, Union

import jax
import jax.numpy as jnp

from flax import nnx

__all__ = ["UNet"]


def timestep_embedding(
    timesteps: jax.Array, dim: int, max_period: int = 10000
) -> jax.Array:
  half = dim // 2
  freqs = jnp.exp(
      -math.log(max_period) *
      jnp.arange(start=0, stop=half, dtype=jnp.float32) / half
  )
  args = timesteps[:, None].astype(jnp.float32) * freqs[None]
  embedding = jnp.concatenate([jnp.cos(args), jnp.sin(args)], axis=-1)
  if dim % 2:
    embedding = jnp.concatenate([embedding,
                                 jnp.zeros_like(embedding[:, :1])],
                                axis=-1)
  return embedding


class GroupNorm32(nnx.GroupNorm):

  def __call__(
      self, x: jax.Array, *, mask: Optional[jax.Array] = None
  ) -> jax.Array:
    return super().__call__(x.astype(jnp.float32), mask=mask).astype(x.dtype)


def conv_nd(
    dims: int,
    in_channels: Union[int, Tuple[int, ...]],
    out_channels: Union[int, Tuple[int, ...]],
    kernel_size: Union[int, Tuple[int, ...]],
    strides: Union[int, Tuple[int, ...]] = 1,
    *,
    dtype: Optional[jnp.dtype] = None,
    param_dtype: jnp.dtype = jnp.float32,
    padding: Union[int, Tuple[int, ...]] = 0,
    zero_init: bool = False,
    rngs: nnx.Rngs,
    **kwargs: Any,
) -> nnx.Conv:
  if isinstance(kernel_size, int):
    kernel_size = (kernel_size,) * dims
  if isinstance(strides, int):
    strides = (strides,) * dims
  if isinstance(padding, int):
    padding = (padding,) * dims

  if zero_init:
    kwargs["kernel_init"] = nnx.initializers.constant(value=0.0)
    kwargs["bias_init"] = nnx.initializers.constant(value=0.0)

  return nnx.Conv(
      in_features=in_channels,
      out_features=out_channels,
      kernel_size=kernel_size,
      strides=strides,
      padding=padding,
      dtype=dtype,
      param_dtype=param_dtype,
      rngs=rngs,
      **kwargs,
  )


def normalization(
    channels: int,
    *,
    dtype: Optional[jnp.dtype] = None,
    param_dtype: jnp.dtype = jnp.float32,
    rngs: nnx.Rngs,
) -> nnx.GroupNorm:
  return GroupNorm32(
      num_groups=32,
      num_features=channels,
      epsilon=1e-5,
      dtype=dtype,
      param_dtype=param_dtype,
      use_fast_variance=False,
      rngs=rngs,
  )


class TimestepBlock(nnx.Module):

  @abc.abstractmethod
  def __call__(
      self,
      x: jax.Array,
      emb: jax.Array,
      *,
      rngs: Optional[nnx.Rngs] = None,
  ) -> jax.Array:
    pass


class TimestepEmbedSequential(nnx.Module):

  def __init__(self, *layers: nnx.Module):
    super().__init__()
    self.layers = nnx.List(layers) if hasattr(nnx, "List") else list(layers)

  def __call__(
      self,
      x: jax.Array,
      emb: Optional[jax.Array] = None,
      *,
      rngs: Optional[nnx.Rngs] = None,
  ) -> jax.Array:
    for layer in self.layers:
      if isinstance(layer, TimestepBlock):
        x = layer(x, emb, rngs=rngs)
      else:
        x = layer(x)
    return x


class Upsample(nnx.Module):

  def __init__(
      self,
      channels: int,
      use_conv: bool,
      *,
      out_channels: Optional[int] = None,
      dtype: Optional[jnp.dtype] = None,
      param_dtype: jnp.dtype = jnp.float32,
      rngs: nnx.Rngs,
  ):
    super().__init__()
    self.channels = channels
    self.out_channels = out_channels or channels
    if use_conv:
      self.conv = conv_nd(
          2,
          self.channels,
          self.out_channels,
          3,
          padding=1,
          dtype=dtype,
          param_dtype=param_dtype,
          rngs=rngs,
      )
    else:
      self.conv = None

  def __call__(self, x: jax.Array) -> jax.Array:
    assert x.shape[-1] == self.channels
    b, h, w, c = x.shape
    shape = (b, 2 * h, 2 * w, c)
    x = jax.image.resize(x, shape, method="nearest")
    if self.conv is not None:
      x = self.conv(x)
    return x


class Downsample(nnx.Module):

  def __init__(
      self,
      channels: int,
      use_conv: bool,
      *,
      out_channels: Optional[int] = None,
      dtype: Optional[jnp.dtype] = None,
      param_dtype: jnp.dtype = jnp.float32,
      rngs: nnx.Rngs,
  ):
    super().__init__()
    self.channels = channels
    self.out_channels = out_channels or channels
    if use_conv:
      self.op = conv_nd(
          2,
          self.channels,
          self.out_channels,
          3,
          strides=2,
          padding=1,
          dtype=dtype,
          param_dtype=param_dtype,
          rngs=rngs,
      )
    else:
      self.op = functools.partial(
          nnx.avg_pool, window_shape=(2, 2), strides=(2, 2)
      )

  def __call__(self, x: jax.Array) -> jax.Array:
    assert x.shape[-1] == self.channels
    return self.op(x)


class ResBlock(TimestepBlock):

  def __init__(
      self,
      channels: int,
      emb_channels: int,
      dropout: float,
      *,
      out_channels: Optional[int] = None,
      use_conv: bool = False,
      up: bool = False,
      down: bool = False,
      dtype: Optional[jnp.dtype] = None,
      param_dtype: jnp.dtype = jnp.float32,
      rngs: nnx.Rngs,
  ):
    super().__init__()
    self.channels = channels
    self.emb_channels = emb_channels
    self.dropout = dropout
    self.out_channels = out_channels or channels
    self.use_conv = use_conv
    self.updown = up or down

    self.in_norm = normalization(
        channels, dtype=dtype, param_dtype=param_dtype, rngs=rngs
    )
    self.in_act = nnx.silu
    self.in_conv = conv_nd(
        2,
        channels,
        self.out_channels,
        3,
        padding=1,
        dtype=dtype,
        param_dtype=param_dtype,
        rngs=rngs,
    )

    if up:
      self.h_upd = Upsample(
          channels,
          use_conv=False,
          dtype=dtype,
          param_dtype=param_dtype,
          rngs=rngs,
      )
      self.x_upd = Upsample(
          channels,
          use_conv=False,
          dtype=dtype,
          param_dtype=param_dtype,
          rngs=rngs,
      )
    elif down:
      self.h_upd = Downsample(
          channels,
          use_conv=False,
          dtype=dtype,
          param_dtype=param_dtype,
          rngs=rngs,
      )
      self.x_upd = Downsample(
          channels,
          use_conv=False,
          dtype=dtype,
          param_dtype=param_dtype,
          rngs=rngs,
      )
    else:
      self.h_upd = lambda x: x
      self.x_upd = lambda x: x

    self.emb_act = nnx.silu
    self.emb_layers = nnx.Linear(
        emb_channels,
        self.out_channels,
        dtype=dtype,
        param_dtype=param_dtype,
        rngs=rngs,
    )

    self.out_norm = normalization(
        self.out_channels, dtype=dtype, param_dtype=param_dtype, rngs=rngs
    )
    self.out_act = nnx.silu
    self.out_dropout = nnx.Dropout(rate=dropout)
    self.out_conv = conv_nd(
        2,
        self.out_channels,
        self.out_channels,
        3,
        padding=1,
        zero_init=True,
        dtype=dtype,
        param_dtype=param_dtype,
        rngs=rngs,
    )

    if self.out_channels == channels:
      self.skip_connection = lambda x: x
    elif use_conv:
      self.skip_connection = conv_nd(
          2,
          channels,
          self.out_channels,
          3,
          padding=1,
          dtype=dtype,
          param_dtype=param_dtype,
          rngs=rngs,
      )
    else:
      self.skip_connection = conv_nd(
          2,
          channels,
          self.out_channels,
          1,
          dtype=dtype,
          param_dtype=param_dtype,
          rngs=rngs,
      )

  def __call__(
      self,
      x: jax.Array,
      emb: jax.Array,
      *,
      rngs: Optional[nnx.Rngs] = None,
  ) -> jax.Array:
    if self.updown:
      h = self.in_norm(x)
      h = self.in_act(h)
      h = self.h_upd(h)
      x = self.x_upd(x)
      h = self.in_conv(h)
    else:
      h = self.in_norm(x)
      h = self.in_act(h)
      h = self.in_conv(h)

    emb_out = self.emb_act(emb).astype(h.dtype)
    emb_out = self.emb_layers(emb_out)
    emb_out = emb_out[:, None, None, :]  # [b, 1, 1, t_emb]
    h = h + emb_out
    h = self.out_norm(h)
    h = self.out_act(h)
    h = self.out_dropout(h, rngs=rngs)
    h = self.out_conv(h)

    return self.skip_connection(x) + h


class QKVAttention(nnx.Module):

  def __init__(
      self,
      n_heads: int,
      attn_implementation: Optional[Literal["xla", "cudnn"]] = None,
  ):
    super().__init__()
    self.n_heads = n_heads
    self.attn_implementation = attn_implementation

  def __call__(self, qkv: jax.Array) -> jax.Array:
    bs, length, width = qkv.shape
    head_dim, rest = divmod(width, 3 * self.n_heads)
    assert rest == 0, rest
    scale = 1.0 / math.sqrt(math.sqrt(head_dim))

    # Split into heads and channels
    q, k, v = jnp.split(qkv, 3, axis=-1)
    q = q.reshape(bs, length, self.n_heads, head_dim)
    k = k.reshape(bs, length, self.n_heads, head_dim)
    v = v.reshape(bs, length, self.n_heads, head_dim)

    dtype = jnp.bfloat16 if self.attn_implementation == "cudnn" else q.dtype
    a = jax.nn.dot_product_attention(
        q.astype(dtype),
        k.astype(dtype),
        v.astype(dtype),
        scale=scale,
        implementation=self.attn_implementation,
    ).astype(q.dtype)
    return a.reshape(bs, length, self.n_heads * head_dim)


class AttentionBlock(nnx.Module):

  def __init__(
      self,
      channels: int,
      *,
      num_heads: int = 1,
      attn_implementation: Optional[Literal["xla", "cudnn"]] = None,
      dtype: Optional[jnp.dtype] = None,
      param_dtype: jnp.dtype = jnp.float32,
      rngs: nnx.Rngs,
  ):
    super().__init__()
    self.channels = channels
    self.num_heads = num_heads
    self.norm = normalization(
        channels, dtype=dtype, param_dtype=param_dtype, rngs=rngs
    )
    self.qkv = conv_nd(
        1,
        channels,
        channels * 3,
        1,
        dtype=dtype,
        param_dtype=param_dtype,
        rngs=rngs,
    )

    # split qkv before split heads
    self.attention = QKVAttention(
        self.num_heads, attn_implementation=attn_implementation
    )

    self.proj_out = conv_nd(
        1,
        channels,
        channels,
        1,
        zero_init=True,
        dtype=dtype,
        param_dtype=param_dtype,
        rngs=rngs,
    )

  def __call__(self, x: jax.Array) -> jax.Array:
    # [B, H, W, C]
    b, *spatial, c = x.shape
    x = x.reshape(b, -1, c)  # [B, H * W, C]
    qkv = self.qkv(self.norm(x))
    h = self.attention(qkv)
    h = self.proj_out(h)
    return (x + h).reshape(b, *spatial, c)


class UNet(nnx.Module):
  """UNet model with attention and timestep embedding.

  Args:
    shape: Input shape ``[height, width, channels]``.
    model_channels: Number of model channels.
    num_res_blocks: Number of residual blocks.
    attention_resolutions: Resolutions at which to add self-attention.
    out_channels: Number of output channels. If :obj:`None`, use input channels.
    dropout: Dropout rate.
    channel_mult: Multiplier for ``model_channels`` for each resolution.
    time_embed_dim: Dimensionality of the time embedding.
      If :obj:`None`, use ``4 * model_channels``.
      If :class:`float`, use ``int(time_embed_dim * model_channels)``.
    conv_resample: If :obj:`False`, don't use convolution for upsampling and use
      average pooling for downsampling instead of using convolution.
    num_heads: Number of attention heads for the input and middle blocks.
    num_heads_upsample: Number of attention heads for the output blocks.
      If :obj:`None`, use ``num_heads``.
    resblock_updown: Whether to use residual blocks for up/downsampling.
    num_classes: Number of classes.
    dtype: Data type for computation.
    param_dtype: Data type for parameters.
    attn_implementation: Attention implementation for
      :func:`~jax.nn.dot_product_attention`.
    rngs: Random number generator for initialization.
  """

  def __init__(
      self,
      *,
      shape: Tuple[int, int, int],
      model_channels: int,
      num_res_blocks: int,
      attention_resolutions: Tuple[int, ...],
      out_channels: Optional[int] = None,
      dropout: float = 0.0,
      channel_mult: Tuple[float, ...] = (1, 2, 4, 8),
      time_embed_dim: Optional[Union[int, float]] = None,
      conv_resample: bool = True,
      num_heads: int = 1,
      num_heads_upsample: Optional[int] = None,
      resblock_updown: bool = False,
      num_classes: Optional[int] = None,
      dtype: Optional[jnp.dtype] = None,
      param_dtype: jnp.dtype = jnp.float32,
      attn_implementation: Optional[Literal["xla", "cudnn"]] = None,
      rngs: nnx.Rngs,
  ):
    super().__init__()

    image_size, _, in_channels = shape
    out_channels = out_channels or in_channels
    attention_resolutions = tuple(
        image_size // res for res in attention_resolutions
    )
    num_heads_upsample = num_heads_upsample or num_heads

    self.dtype = dtype
    self.in_channels = in_channels
    self.model_channels = model_channels
    self.num_res_blocks = num_res_blocks
    self.attention_resolutions = attention_resolutions
    self.dropout = dropout
    self.channel_mult = channel_mult
    self.conv_resample = conv_resample
    self.num_heads = num_heads
    self.num_heads_upsample = num_heads_upsample

    # Time embedding
    if time_embed_dim is None:
      time_embed_dim = model_channels * 4
    elif isinstance(time_embed_dim, float):
      time_embed_dim = int(time_embed_dim * model_channels)
    assert isinstance(time_embed_dim, int), time_embed_dim

    self.time_embed = TimestepEmbedSequential(
        nnx.Linear(
            model_channels,
            time_embed_dim,
            dtype=dtype,
            param_dtype=param_dtype,
            rngs=rngs,
        ),
        nnx.silu,
        nnx.Linear(
            time_embed_dim,
            time_embed_dim,
            dtype=dtype,
            param_dtype=param_dtype,
            rngs=rngs,
        ),
    )

    # condition embedding
    if num_classes is not None:
      self.label_emb = nnx.Embed(
          num_embeddings=num_classes, features=time_embed_dim, rngs=rngs
      )
    else:
      self.label_emb = None

    # Input blocks
    self.input_blocks = nnx.List() if hasattr(nnx, "List") else []
    input_block_chans = [model_channels]
    ch = int(channel_mult[0] * model_channels)
    ds = 1

    # First input block (just convolution)
    self.input_blocks.append(
        TimestepEmbedSequential(
            conv_nd(
                2,
                in_channels,
                ch,
                3,
                padding=1,
                dtype=dtype,
                param_dtype=param_dtype,
                rngs=rngs,
            )
        )
    )

    # Rest of input blocks
    for level, mult in enumerate(channel_mult):
      for _ in range(num_res_blocks):
        layers = [
            ResBlock(
                ch,
                time_embed_dim,
                dropout,
                out_channels=int(mult * model_channels),
                dtype=dtype,
                param_dtype=param_dtype,
                rngs=rngs,
            )
        ]
        ch = int(mult * model_channels)

        if ds in attention_resolutions:
          layers.append(
              AttentionBlock(
                  ch,
                  num_heads=num_heads,
                  attn_implementation=attn_implementation,
                  dtype=dtype,
                  param_dtype=param_dtype,
                  rngs=rngs,
              )
          )

        self.input_blocks.append(TimestepEmbedSequential(*layers))
        input_block_chans.append(ch)

      # Downsample if not last level
      if level != len(channel_mult) - 1:
        out_ch = ch
        self.input_blocks.append(
            TimestepEmbedSequential(
                ResBlock(
                    ch,
                    time_embed_dim,
                    dropout,
                    out_channels=out_ch,
                    down=True,
                    dtype=dtype,
                    param_dtype=param_dtype,
                    rngs=rngs,
                ) if resblock_updown else Downsample(
                    ch,
                    conv_resample,
                    out_channels=out_ch,
                    dtype=dtype,
                    param_dtype=param_dtype,
                    rngs=rngs,
                )
            )
        )
        ch = out_ch
        input_block_chans.append(ch)
        ds *= 2

    self.middle_block = TimestepEmbedSequential(
        ResBlock(
            ch,
            time_embed_dim,
            dropout,
            dtype=dtype,
            param_dtype=param_dtype,
            rngs=rngs,
        ),
        AttentionBlock(
            ch,
            num_heads=num_heads,
            attn_implementation=attn_implementation,
            dtype=dtype,
            param_dtype=param_dtype,
            rngs=rngs,
        ),
        ResBlock(
            ch,
            time_embed_dim,
            dropout,
            dtype=dtype,
            param_dtype=param_dtype,
            rngs=rngs,
        ),
    )

    # Output blocks
    self.output_blocks = nnx.List() if hasattr(nnx, "List") else []
    for level, mult in list(enumerate(channel_mult))[::-1]:
      for i in range(num_res_blocks + 1):
        ich = input_block_chans.pop()
        layers = [
            ResBlock(
                ch + ich,
                time_embed_dim,
                dropout,
                out_channels=int(model_channels * mult),
                dtype=dtype,
                param_dtype=param_dtype,
                rngs=rngs,
            )
        ]
        ch = int(model_channels * mult)

        if ds in attention_resolutions:
          layers.append(
              AttentionBlock(
                  ch,
                  num_heads=num_heads_upsample,
                  attn_implementation=attn_implementation,
                  dtype=dtype,
                  param_dtype=param_dtype,
                  rngs=rngs,
              )
          )

        if level and i == num_res_blocks:
          out_ch = ch
          layers.append(
              ResBlock(
                  ch,
                  time_embed_dim,
                  dropout,
                  out_channels=out_ch,
                  up=True,
                  dtype=dtype,
                  param_dtype=param_dtype,
                  rngs=rngs,
              ) if resblock_updown else Upsample(
                  ch,
                  conv_resample,
                  out_channels=out_ch,
                  dtype=dtype,
                  param_dtype=param_dtype,
                  rngs=rngs,
              )
          )
          ds //= 2

        self.output_blocks.append(TimestepEmbedSequential(*layers))

    self.out = TimestepEmbedSequential(
        normalization(ch, dtype=dtype, param_dtype=param_dtype, rngs=rngs),
        nnx.silu,
        conv_nd(
            2,
            ch,
            out_channels,
            3,
            padding=1,
            zero_init=True,
            # dtype=dtype,  # don't cast computations to (bf)float16
            param_dtype=param_dtype,
            rngs=rngs,
        ),
    )

  def __call__(
      self,
      t: jax.Array,
      x: jax.Array,
      cond: Optional[jax.Array] = None,
      *,
      rngs: Optional[nnx.Rngs] = None,
  ) -> jax.Array:
    """Compute the velocity.

    Args:
      t: Time array of shape ``[batch,]``.
      x: Image of shape ``[batch, height, width, channels]``.
      cond: Class condition array of shape ``[batch,]``.
      rngs: Random number generator for dropout.

    Returns:
      The velocity array of shape ``[batch, height, width, channels]``.
    """
    emb = self.time_embed(timestep_embedding(t, self.model_channels), rngs=rngs)
    # TODO(michalk8): generalize for different types of conditions
    if self.label_emb is not None:
      assert cond is not None, "Please provide a condition."
      # emb is cast to `self.dtype` inside each submodule
      emb = emb + self.label_emb(cond)
    h = x.astype(self.dtype)
    hs = []
    for module in self.input_blocks:
      h = module(h, emb, rngs=rngs)
      hs.append(h)
    h = self.middle_block(h, emb, rngs=rngs)
    for module in self.output_blocks:
      h = jnp.concatenate([h, hs.pop()], axis=-1)
      h = module(h, emb, rngs=rngs)
    h = h.astype(x.dtype)
    # output's compute dtype
    return self.out(h, rngs=rngs)



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
from typing import Any, Callable, Optional, Sequence

import jax
from jax import numpy as jnp

from flax import nnx

__all__ = ["MLP"]


class MLP(nnx.Module):
  """MLP velocity field.

  Args:
    dim: Dimensionality of the velocity field.
    hidden_dims: Hidden dimensions.
    cond_dim: Dimensionality of the condition vector.
    act_fn: Activation function.
    dropout_rate: Dropout rate.
    rngs: Random number generator used for initialization.
    kwargs: Keyword arguments for :class:`~flax.nnx.Linear`.
  """

  def __init__(
      self,
      dim: int,
      *,
      hidden_dims: Sequence[int] = (),
      cond_dim: int = 0,
      act_fn: Callable[[jax.Array], jax.Array] = nnx.silu,
      time_enc_num_freqs: Optional[int] = None,
      dropout_rate: float = 0.0,
      rngs: nnx.Rngs,
      **kwargs: Any,
  ):
    super().__init__()
    if time_enc_num_freqs is None:
      time_enc_num_freqs = max(1, min(dim // 16, 64))

    time_dim = 2 * time_enc_num_freqs
    hidden_dims = [dim + time_dim + cond_dim] + list(hidden_dims)
    block_fn = functools.partial(
        Block, act_fn=act_fn, dropout_rate=dropout_rate, **kwargs
    )

    blocks = [
        *(
            block_fn(in_dim, out_dim, rngs=rngs)
            for in_dim, out_dim in zip(hidden_dims[:-1], hidden_dims[1:])
        )
    ]
    final_block = block_fn(
        hidden_dims[-1],
        dim,
        act_fn=jax.nn.identity,
        rngs=rngs,
    )

    self.blocks = nnx.Sequential(*blocks, final_block)
    self.time_enc_num_freqs = time_enc_num_freqs

  def __call__(
      self,
      t: jax.Array,
      x: jax.Array,
      cond: Optional[jax.Array] = None,
      *,
      rngs: Optional[nnx.Rngs] = None,
  ) -> jax.Array:
    """Compute the velocity.

    Args:
      t: Time array of shape ``[batch,]``.
      x: Input array of shape ``[batch, dim]``.
      cond: Condition array of shape ``[batch, cond_dim]``.
      rngs: Random number generator for dropout.

    Returns:
      The velocity array of shape ``[batch, dim]``.
    """
    t_emb = _encode_time(t, self.time_enc_num_freqs)
    h = [t_emb, x] if cond is None else [t_emb, x, cond]
    h = jnp.concatenate(h, axis=-1)
    return self.blocks(h, rngs=rngs)


class Block(nnx.Module):

  def __init__(
      self,
      in_dim: int,
      out_dim: int,
      *,
      act_fn: Callable[[jax.Array], jax.Array],
      dropout_rate: float = 0.0,
      rngs: nnx.Rngs,
      **kwargs: Any
  ):
    super().__init__()
    self.lin = nnx.Linear(in_dim, out_dim, rngs=rngs, **kwargs)
    self.act_fn = act_fn
    self.dropout = nnx.Dropout(dropout_rate)

  def __call__(
      self, x: jax.Array, *, rngs: Optional[nnx.Rngs] = None
  ) -> jax.Array:
    return self.dropout(self.act_fn(self.lin(x)), rngs=rngs)


def _encode_time(t: jax.Array, num_freqs: int) -> jax.Array:
  assert num_freqs > 0, "Number of frequencies must be positive."
  assert t.ndim == 1, t.shape
  freq = (2 * jnp.pi) * jnp.arange(1, num_freqs + 1)
  t = freq * t[:, None]
  return jnp.concatenate([jnp.cos(t), jnp.sin(t)], axis=-1)



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
from typing import Any, NamedTuple, Optional, Tuple, Union

import jax
import jax.numpy as jnp

from ott import utils
from ott.geometry import pointcloud
from ott.math import fixed_point_loop
from ott.math import utils as mu
from ott.problems.linear import barycenter_problem, linear_problem
from ott.solvers import was_solver
from ott.solvers.linear import sinkhorn, sinkhorn_lr

__all__ = ["FreeBarycenterState", "FreeWassersteinBarycenter"]

LinearOutput = Union[sinkhorn.SinkhornOutput, sinkhorn_lr.LRSinkhornOutput]


class FreeBarycenterState(NamedTuple):
  """Holds the state of the Wasserstein barycenter solver.

  Args:
    x: Barycenter points.
    a: Barycenter weights.
    costs: Holds the sequence of regularized costs seen through the outer
      loop of the solver.
    linear_convergence: Holds the sequence of bool convergence flags of the
      inner Sinkhorn iterations.
    linear_outputs: Holds latest output objects returned by linear solver when
      recomputing transport between barycenter vs. each of the input measures.
    errors: Holds sequence of vectors of errors of the Sinkhorn algorithm
      at each iteration.
  """

  x: jnp.ndarray
  a: jnp.ndarray
  costs: Optional[jnp.ndarray] = None
  linear_convergence: Optional[jnp.ndarray] = None
  linear_outputs: Optional[LinearOutput] = None
  errors: Optional[jnp.ndarray] = None

  def set(self, **kwargs: Any) -> "FreeBarycenterState":
    """Return a copy of self, possibly with overwrites."""
    return self._replace(**kwargs)

  def update(
      self, iteration: int, bar_prob: barycenter_problem.FreeBarycenterProblem,
      linear_solver: Any, store_errors: bool
  ) -> "FreeBarycenterState":
    """Update the state of the solver.

    Args:
      iteration: the current iteration of the outer loop.
      bar_prob: the barycenter problem.
      linear_solver: the linear OT solver to use.
      store_errors: whether to store the errors of the inner loop.

    Returns:
      The updated state.
    """
    seg_y, seg_b, _ = bar_prob.segmented_y_b

    @functools.partial(jax.vmap, in_axes=[None, None, 0, 0])
    def solve_linear_ot(
        a: Optional[jnp.ndarray], x: jnp.ndarray, b: jnp.ndarray, y: jnp.ndarray
    ):
      geom = pointcloud.PointCloud(
          x, y, cost_fn=bar_prob.cost_fn, epsilon=bar_prob.epsilon
      )
      prob = linear_problem.LinearProblem(geom, a=a, b=b)
      out = linear_solver(prob)
      # instantiate matrix since it is a property of out.
      return out, out.matrix

    outs, matrices = solve_linear_ot(self.a, self.x, seg_b, seg_y)
    reg_ot_costs = outs.reg_ot_cost
    convergeds = outs.converged
    errors = outs.errors

    cost = jnp.sum(reg_ot_costs * bar_prob.weights)
    updated_costs = self.costs.at[iteration].set(cost)
    converged = jnp.all(convergeds)
    linear_convergence = self.linear_convergence.at[iteration].set(converged)

    if store_errors and self.errors is not None:
      errors = self.errors.at[iteration, :, :].set(errors)
    else:
      errors = None

    # Approximation of barycenter as barycenter of barycenters per measure.

    barycenters_per_measure = mu.barycentric_projection(
        matrices, seg_y, bar_prob.cost_fn
    )

    x_new = jax.vmap(
        lambda w, y: bar_prob.cost_fn.barycenter(w, y)[0], in_axes=[None, 1]
    )(bar_prob.weights, barycenters_per_measure)

    return self.set(
        x=x_new,
        costs=updated_costs,
        linear_convergence=linear_convergence,
        linear_outputs=outs,
        errors=errors
    )


class FreeBarycenterOutput(NamedTuple):
  """Holds the output of a Free Wasserstein barycenter solver.

  Objects of this class contain both solutions and problem definition of a
  regularized Free Barycenter OT problem, along several methods that can be used
  to access its content.

  Args:
    x: barycenter points.
    a: barycenter weights.
    costs: Holds the sequence of regularized GW costs seen through the outer
      loop of the solver.
    linear_convergence: Holds the sequence of bool convergence flags of the
      inner Sinkhorn iterations.
    linear_outputs: Holds latest output objects returned by linear solver when
      recomputing transport between barycenter vs. each of the input measures.
    errors: Holds sequence of vectors of errors of the Sinkhorn algorithm
      at each iteration.
  """

  x: jnp.ndarray
  a: jnp.ndarray
  bar_prob: barycenter_problem.FreeBarycenterProblem
  costs: jnp.ndarray
  linear_convergence: jnp.ndarray
  linear_outputs: LinearOutput
  errors: Optional[jnp.ndarray] = None

  @property
  def all_linear_solvers_converged(self) -> bool:
    """Whether all linear convergence flags converged."""
    return jnp.all(self.linear_convergence[self.linear_convergence != -1])

  def matrix_at_index(self, measure_index: int) -> jnp.ndarray:
    """Return the transport matrix from barycenter to measure_index measure."""
    size_measure = self.bar_prob.num_per_measure[measure_index]
    matrix = self.linear_output_at_index(measure_index).matrix
    return jax.lax.dynamic_slice_in_dim(matrix, 0, size_measure, axis=-1)

  @property
  def bar_size(self) -> int:
    """Size of the barycenter."""
    return self.x.shape[0]

  @property
  def num_iters(self) -> int:
    """Number of outer iterations performed to converge."""
    return jnp.sum(self.linear_convergence != -1)

  @property
  def costs_along_iterations(self) -> jnp.ndarray:
    """Costs vector with superfluous values removed."""
    return self.costs[:self.num_iters]

  def linear_output_at_index(self, i: int) -> LinearOutput:
    """Linear solver output of transport from barycenter to measure i."""
    return jax.tree.map(lambda x: x[i], self.linear_outputs)


@jax.tree_util.register_pytree_node_class
class FreeWassersteinBarycenter(was_solver.WassersteinSolver):
  """Continuous Wasserstein barycenter solver :cite:`cuturi:14`."""

  def __call__(  # noqa: D102
      self,
      bar_prob: barycenter_problem.FreeBarycenterProblem,
      bar_size: int = 100,
      x_init: Optional[jnp.ndarray] = None,
      rng: Optional[jax.Array] = None,
  ) -> FreeBarycenterState:
    rng = utils.default_prng_key(rng)
    return self.iterations(bar_size, bar_prob, x_init, rng)

  def init_state(
      self,
      bar_prob: barycenter_problem.FreeBarycenterProblem,
      bar_size: int,
      x_init: Optional[jnp.ndarray] = None,
      rng: Optional[jax.Array] = None,
  ) -> FreeBarycenterState:
    """Initialize the state of the Wasserstein barycenter iterations.

    Args:
      bar_prob: The barycenter problem.
      bar_size: Size of the barycenter.
      x_init: Initial barycenter estimate of shape ``[bar_size, ndim]``.
        If `None`, ``bar_size`` points will be sampled from the input
        measures according to their weights
        :attr:`~ott.problems.linear.barycenter_problem.FreeBarycenterProblem.flattened_y`.
      rng: Random key for seeding.

    Returns:
      The initial barycenter state.
    """
    if x_init is not None:
      assert bar_size == x_init.shape[0]
      x = x_init
    else:
      # sample randomly points in the support of the y measures
      rng = utils.default_prng_key(rng)
      indices_subset = jax.random.choice(
          rng,
          a=bar_prob.flattened_y.shape[0],
          shape=(bar_size,),
          replace=False,
          p=bar_prob.flattened_b
      )
      x = bar_prob.flattened_y[indices_subset, :]

    # TODO(cuturi) expand to non-uniform weights for barycenter.
    a = jnp.ones((bar_size,)) / bar_size
    num_iter = self.max_iterations
    if self.store_inner_errors:
      errors = -jnp.ones((
          num_iter, bar_prob.num_measures, self.linear_solver.outer_iterations
      ))
    else:
      errors = None
    state = FreeBarycenterState(
        x=x,
        a=a,
        costs=-jnp.ones((num_iter,)),
        linear_convergence=-jnp.ones((num_iter,)),
        errors=errors
    )
    abstract_tree = jax.eval_shape(
        functools.partial(state.update, store_errors=self.store_inner_errors),
        0, bar_prob, self.linear_solver
    )

    linear_outputs = jax.tree.map(jnp.zeros_like, abstract_tree.linear_outputs)

    return FreeBarycenterState(
        x=x,
        a=a,
        costs=-jnp.ones((num_iter,)),
        linear_convergence=-jnp.ones((num_iter,)),
        linear_outputs=linear_outputs,
        errors=errors
    )

  def output_from_state(  # noqa: D102
      self,
      state: FreeBarycenterState,
      bar_prob: barycenter_problem.FreeBarycenterProblem
  ) -> FreeBarycenterOutput:
    """Create an output from a barycenter state."""
    return FreeBarycenterOutput(
        x=state.x,
        a=state.a,
        bar_prob=bar_prob,
        costs=state.costs,
        linear_convergence=state.linear_convergence,
        linear_outputs=state.linear_outputs,
        errors=state.errors,
    )

  def iterations(
      self, bar_size: int, bar_prob: barycenter_problem.FreeBarycenterProblem,
      x_init: jnp.ndarray, rng: jax.Array
  ) -> FreeBarycenterState:
    """Wasserstein barycenter outer loop."""

    def cond_fn(
        iteration: int,
        constants: Tuple[FreeWassersteinBarycenter,
                         barycenter_problem.FreeBarycenterProblem],
        state: FreeBarycenterState
    ) -> bool:
      return self._continue(state, iteration)

    def body_fn(
        iteration, constants: Tuple[FreeWassersteinBarycenter,
                                    barycenter_problem.FreeBarycenterProblem],
        state: FreeBarycenterState, compute_error: bool
    ) -> FreeBarycenterState:
      del compute_error  # Always assumed True
      bar_prob = constants
      return state.update(
          iteration, bar_prob, self.linear_solver, self.store_inner_errors
      )

    state = fixed_point_loop.fixpoint_iter(
        cond_fn=cond_fn,
        body_fn=body_fn,
        min_iterations=self.min_iterations,
        max_iterations=self.max_iterations,
        inner_iterations=1,
        constants=bar_prob,
        state=self.init_state(bar_prob, bar_size, x_init, rng)
    )

    return self.output_from_state(state, bar_prob)



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import math
import operator
from typing import Any, Callable, Optional, Tuple

import jax
import jax.numpy as jnp
import jax.tree_util as jtu
import numpy as np
import scipy as sp
from scipy.stats import qmc

from ott import utils
from ott.geometry import costs, pointcloud
from ott.solvers import linear
from ott.solvers.linear import sinkhorn

__all__ = ["OTCP", "sobol_ball_sampler"]

ScoreFn = Callable[[jnp.ndarray, jnp.ndarray], jnp.ndarray]


def sobol_ball_sampler(
    rng: Optional[jax.Array],
    shape: Tuple[int, int],
    n_per_radius: Optional[int] = None,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
  """Sample target measure for :class:`OTCP`.

  Args:
    rng: Random number generator.
    shape: Tuple of ``[n_samples, dim]``.
    n_per_radius: Optionally specify how many samples to sample per radius.

  Returns:
    Points of shape ``[n_S * n_R, dim]`` and weights ``[n_S * n_R,]``.
  """
  rng = utils.default_prng_key(rng)

  n_samples, dim = shape
  n_per_radius = math.ceil(
      math.sqrt(n_samples)
  ) if n_per_radius is None else n_per_radius
  n_sphere, n_0s = divmod(n_samples, n_per_radius)

  radius_grid = jnp.linspace(1.0 / n_per_radius, 1.0, n_per_radius)

  seed = jax.random.randint(rng, shape=(), minval=0, maxval=2 ** 16 - 1)
  out_struct = jax.ShapeDtypeStruct(
      shape=(n_sphere, dim), dtype=radius_grid.dtype
  )
  sphere = jax.pure_callback(_sobol_sphere, out_struct, n_sphere, dim, seed)

  points = sphere[None] * radius_grid[:, None, None]
  points = points.reshape(-1, dim)

  weights = jnp.full(points.shape[0] + (n_0s > 0), fill_value=1.0 / n_samples)
  if n_0s:
    points = jnp.vstack([points, jnp.zeros([1, dim])])
    weights = weights.at[-1].set(n_0s / n_samples)
  return points, weights


def _sobol_sphere(n: int, d: int, seed: int) -> np.ndarray:
  # convert because usually called from the `pure_callback`
  n, d, seed = int(n), int(d), int(seed)
  sampler = qmc.Sobol(d=d, seed=seed, scramble=True)
  points = sampler.random_base2(m=math.ceil(math.log2(n)))[:n]
  points = sp.special.ndtri(points)
  return points / (
      np.linalg.norm(points, keepdims=True, axis=-1) +
      np.finfo(points.dtype).tiny
  )


@jtu.register_dataclass
@dataclasses.dataclass(frozen=True)
class OTCP:
  """Optimal transport conformal prediction :cite:`klein:25`.

  Args:
    model: Fitted model.
    nonconformity_fn: Multivariate nonconformity score function with a signature
      ``(target, prediction) -> score``.
    sinkhorn_output: Sinkhorn output computed in :meth:`fit_transport`.
    sampler: sampler function used to sample points from a reference measure.
      :func:`~ott.tools.conformal.sobol_ball_sampler` is used by default.
    offset: Offset used when re-scaling the data.
    scale: Scale when re-scaling the data.
    calibration_scores: Nonconformity calibration scores computed in
      :meth:`calibrate`.
  """
  model: Callable[[jnp.ndarray],
                  jnp.ndarray] = dataclasses.field(metadata={"static": True})
  nonconformity_fn: ScoreFn = dataclasses.field(
      default=operator.sub, metadata={"static": True}
  )
  sinkhorn_output: Optional[sinkhorn.SinkhornOutput] = None
  sampler: Optional[Callable[[jax.random.PRNGKey, Tuple[int, int]],
                             jax.Array]] = dataclasses.field(
                                 default=None, metadata={"static": True}
                             )
  offset: jnp.ndarray = 0.0
  scale: jnp.ndarray = 1.0
  calibration_scores: Optional[jnp.ndarray] = None

  def fit_transport(
      self,
      x: jnp.ndarray,
      y: jnp.ndarray,
      epsilon: float = 1e-1,
      n_target: int = 8192,
      rng: Optional[jax.Array] = None,
      sampler_kwargs: Optional[Any] = None,
      **kwargs: Any,
  ) -> "OTCP":
    """Fit the transport map.

    Args:
      x: Features of shape ``[n, dim_x]`` to fit the transport map.
      y: Targets of shape ``[n, dim_y]`` to fit the transport map.
      epsilon: Epsilon regularization
      n_target: Total number of points used to create the target measure.
      rng: Random number generator.
      sampler_kwargs: Keyword arguments passed for the :attr:`sampler`.
      kwargs: Keyword arguments for :func:`~ott.solvers.linear.solve`.

    Returns:
      Self and updates :attr:`sinkhorn_output`, :attr:`offset`, :attr:`scale`.
    """
    assert y.ndim == 2, y.shape
    dim = y.shape[-1]

    y_hat_trn = self.model(x)
    scores = self.nonconformity_fn(y, y_hat_trn)
    offset = jnp.mean(scores, axis=0, keepdims=True)
    scale = jnp.linalg.norm(scores - offset, axis=-1).max()
    scores = (scores - offset) / scale

    sampler_kwargs = sampler_kwargs or {}
    sampler = sobol_ball_sampler if self.sampler is None else self.sampler

    target, weights = sampler(rng=rng, shape=(n_target, dim), **sampler_kwargs)
    geom = pointcloud.PointCloud(
        scores, target, epsilon=epsilon, cost_fn=costs.SqEuclidean()
    )
    out = linear.solve(geom, b=weights, **kwargs)

    return dataclasses.replace(
        self, sinkhorn_output=out, offset=offset, scale=scale
    )

  def calibrate(self, x: jnp.ndarray, y: jnp.ndarray) -> "OTCP":
    """Compute calibration scores.

    Args:
      x: Calibration features of shape ``[n_calib, dim_x]``.
      y: Calibration targets of shape ``[n_calib, dim_y]``.

    Returns:
      Self and updates :attr:`calibration_scores`.
    """
    scores = self.get_scores(x, y)
    return dataclasses.replace(self, calibration_scores=scores)

  def get_scores(self, x: jnp.ndarray, y: jnp.ndarray) -> jnp.ndarray:
    """Compute nonconformity scores.

    Args:
      x: Features of shape ``[n, dim_x]``.
      y: Targets of shape ``[n, dim_y]``.

    Returns:
      Nonconformity scores of shape ``[n,]``.
    """
    return self._get_scores(y, self.model(x))

  def predict(
      self,
      x: jnp.ndarray,
      y_candidates: Optional[jnp.ndarray] = None,
      alpha: float = 0.1,
  ) -> jnp.ndarray:
    """Conformalize the model's prediction.

    Args:
      x: Features of shape ``[..., dim_x]``.
      y_candidates: Candidate targets of shape ``[m, dim_y]``.
      alpha: Miscoverage level.

    Returns:
      If ``y_candidates = None``, return an array of shape
      ``[..., n_target, dim_y]``, else a boolean array of shape ``[..., m]``.
    """
    assert x.ndim in (1, 2), x.shape
    assert self.calibration_scores is not None, "Run `.calibrate()` first."

    y_hat = self.model(jnp.atleast_2d(x))
    quantile = jnp.quantile(self.calibration_scores, q=1 - alpha)
    if y_candidates is None:
      res = self._predict_backward(y_hat, quantile=quantile)
    else:
      res = self._predict_forward(y_hat, y_candidates, quantile=quantile)
    return res.squeeze(0) if x.ndim == 1 else res

  def _predict_backward(
      self,
      y_hat: jnp.ndarray,
      *,
      quantile: float,
  ) -> jnp.ndarray:
    candidates = self._transport(quantile * self.target_measure, forward=False)
    candidates = self._rescale(candidates, forward=False)
    return y_hat[:, None] + candidates[None]

  def _predict_forward(
      self,
      y_hat: jnp.ndarray,
      y_candidates: jnp.ndarray,
      *,
      quantile: float,
  ) -> jnp.ndarray:
    assert y_candidates.ndim == 2, y_candidates.shape
    score_fn = jax.vmap(
        jax.vmap(self._get_scores, in_axes=[0, None]), in_axes=[None, 0]
    )
    scores = score_fn(y_candidates, y_hat)
    return scores <= quantile

  def _get_scores(self, y: jnp.ndarray, y_hat: jnp.ndarray) -> jnp.ndarray:
    scores = self.nonconformity_fn(jnp.atleast_2d(y), jnp.atleast_2d(y_hat))
    scores = self._rescale(scores, forward=True)
    scores = self._transport(scores, forward=True)
    scores = jnp.linalg.norm(scores, axis=-1)
    return scores.squeeze(0) if y.ndim == 1 else scores

  def _transport(self, x: jnp.ndarray, *, forward: bool = True) -> jnp.ndarray:
    assert self.sinkhorn_output is not None, "Run `.fit_transport()` first."
    return self.sinkhorn_output.to_dual_potentials().transport(
        x, forward=forward
    )

  def _rescale(self, x: jnp.ndarray, *, forward: bool) -> jnp.ndarray:
    if forward:
      return (x - self.offset) / self.scale
    return (self.scale * x) + self.offset

  @property
  def target_measure(self) -> Optional[jnp.ndarray]:
    """Target measure of shape ``[n_target, dim_y]``."""
    return None if self.sinkhorn_output is None else self.sinkhorn_output.geom.y



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import functools
from typing import Any, Iterable, Iterator, Literal, Optional, Tuple

import jax
import jax.random as jr

from ott.geometry import costs, pointcloud
from ott.solvers import linear

__all__ = ["LinearOTDataloader"]


@dataclasses.dataclass(frozen=False, repr=False)
class LinearOTDataloader:
  """Linear OT dataloader.

  This dataloader wraps a dataloader that generates ``(source, target)``
  arrays with shape ``[batch, ...]`` and aligns them
  using the :class:`~ott.solvers.linear.sinkhorn.Sinkhorn` algorithm.

  Args:
    rng: Random number generator.
    dataset: Iterable dataset which yields a tuple of source and target arrays
      of shape ``[batch, ...]``.
    epsilon: Epsilon regularization.
      See :class:`~ott.geometry.geometry.Geometry` for more information.
    relative_epsilon: Whether ``epsilon`` refers to a fraction of the
      :attr:`~ott.geometry.pointcloud.PointCloud.mean_cost_matrix` or
      :attr:`~ott.geometry.pointcloud.PointCloud.std_cost_matrix`.
    cost_fn: Cost function between two points.
    threshold: Convergence threshold for
      :class:`~ott.solvers.linear.sinkhorn.Sinkhorn`.
    max_iterations: Maximum number of Sinkhorn iterations.
    replace: Whether to sample with replacement.
    shardings: Input and output shardings for the source and target arrays.
  """
  rng: jax.Array
  dataset: Iterable[Tuple[jax.Array, jax.Array]]
  epsilon: Optional[float] = None
  relative_epsilon: Optional[Literal["mean", "std"]] = None
  cost_fn: Optional[costs.CostFn] = None
  threshold: float = 1e-3
  max_iterations: int = 2000
  replace: bool = True
  shardings: Optional[jax.sharding.Sharding] = None

  def __post_init__(self) -> None:
    self._align_fn = jax.jit(
        functools.partial(
            _align,
            threshold=self.threshold,
            max_iterations=self.max_iterations
        ),
        static_argnames=["cost_fn", "epsilon", "relative_epsilon", "replace"],
        in_shardings=(None, self.shardings, self.shardings),
        out_shardings=(self.shardings, self.shardings),
    )
    self._data_it: Optional[Iterator[Tuple[jax.Array, jax.Array]]] = None
    self._rng_it: Optional[jax.Array] = None

  def __iter__(self) -> "LinearOTDataloader":
    """Return self."""
    self._data_it = iter(self.dataset)
    self._rng_it = self.rng
    return self

  def __next__(self) -> Tuple[jax.Array, jax.Array]:
    """Align source and target samples in a batch.

    Returns:
      The aligned source and target arrays of shape ``[batch, ...]``.
    """
    assert self._data_it is not None, "Please call `iter()` first."
    assert self._rng_it is not None, "Please call `iter()` first."
    self._rng_it, rng_sample = jr.split(self._rng_it, 2)
    x, y = next(self._data_it)
    return self._align_fn(
        rng_sample,
        x,
        y,
        self.cost_fn,
        self.epsilon,
        self.relative_epsilon,
        self.replace,
    )


def _align(
    rng: jax.Array,
    x: jax.Array,
    y: jax.Array,
    cost_fn: costs.CostFn,
    epsilon: Optional[float],
    relative_epsilon: Optional[Literal["mean", "std"]],
    replace: bool,
    **kwargs: Any,
) -> Tuple[jax.Array, jax.Array]:
  geom = pointcloud.PointCloud(
      x,
      y,
      cost_fn=cost_fn,
      epsilon=epsilon,
      relative_epsilon=relative_epsilon,
  )
  out = linear.solve(geom, **kwargs)

  n, m = geom.shape
  probs = out.matrix.ravel()
  probs = probs / probs.sum()

  ixs = jr.choice(rng, n * m, shape=(n,), p=probs, replace=replace)
  row_ixs, col_ixs = ixs // m, ixs % m
  return x[row_ixs], y[col_ixs]



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
from typing import TYPE_CHECKING, Any, Callable, Literal, Optional, Tuple, Union

if TYPE_CHECKING:
  from ott.geometry import low_rank

import jax
import jax.numpy as jnp
import jax.scipy as jsp
import jax.tree_util as jtu

from ott import utils
from ott.geometry import epsilon_scheduler as eps_scheduler
from ott.math import utils as mu

__all__ = ["Geometry"]


@jtu.register_pytree_node_class
class Geometry:
  r"""Base class to define ground costs/kernels used in optimal transport.

  Optimal transport problems are intrinsically geometric: they compute an
  optimal way to transport mass from one configuration onto another. To define
  what is meant by optimality of transport requires defining a
  :term:`ground cost`, which quantifies how costly it is to move mass from
  one among several source locations, towards one out of multiple
  target locations. These source and target locations can be described as
  points in vectors spaces, grids, or more generally described
  through a (dissimilarity) cost matrix, or almost equivalently, a
  (similarity) kernel matrix. This class describes such a
  geometry and several useful methods to exploit it.

  Args:
    cost_matrix: Cost matrix of shape ``[n, m]``.
    kernel_matrix: Kernel matrix of shape ``[n, m]``.
    epsilon: Regularization parameter or a scheduler:

      - ``epsilon = None`` and ``relative_epsilon = None``, use
        :math:`0.05 * \text{stddev(cost_matrix)}`.
      - if ``epsilon`` is a :class:`float` and ``relative_epsilon = None``,
        it directly corresponds to the regularization strength.
      - otherwise, ``epsilon`` multiplies the :attr:`mean_cost_matrix` or
        :attr:`std_cost_matrix`, depending on the value of ``relative_epsilon``.

      If ``epsilon = None``, the value of
      :obj:`DEFAULT_EPSILON_SCALE = 0.05 <ott.geometry.epsilon_scheduler.DEFAULT_EPSILON_SCALE>`.
      will be used.
    relative_epsilon: Whether ``epsilon`` refers to a fraction of the
      :attr:`mean_cost_matrix` or :attr:`std_cost_matrix`.
    scale_cost: option to rescale the cost matrix. Implemented scalings are
      'median', 'mean', 'std' and 'max_cost'. Alternatively, a float factor can
      be given to rescale the cost such that ``cost_matrix /= scale_cost``.

  Note:
    When defining a :class:`~ott.geometry.geometry.Geometry` through a
    ``cost_matrix``, it is important to select an ``epsilon`` regularization
    parameter that is meaningful. That parameter can be provided by the user,
    or assigned a default value through a simple rule, using for instance the
    :attr:`mean_cost_matrix` or the :attr:`std_cost_matrix`.
  """  # noqa: E501

  def __init__(
      self,
      cost_matrix: Optional[jnp.ndarray] = None,
      kernel_matrix: Optional[jnp.ndarray] = None,
      epsilon: Optional[Union[float, eps_scheduler.Epsilon]] = None,
      relative_epsilon: Optional[Literal["mean", "std"]] = None,
      scale_cost: Union[float, Literal["mean", "max_cost", "median",
                                       "std"]] = 1.0,
  ):
    self._cost_matrix = cost_matrix
    self._kernel_matrix = kernel_matrix
    self._epsilon_init = epsilon
    self._relative_epsilon = relative_epsilon
    self._scale_cost = scale_cost

  @property
  def cost_rank(self) -> Optional[int]:
    """Output rank of cost matrix, if any was provided."""

  @property
  def cost_matrix(self) -> jnp.ndarray:
    """Cost matrix, recomputed from kernel if only kernel was specified."""
    if self._cost_matrix is None:
      # If no epsilon was passed on to the geometry, then assume it is one by
      # default.
      eps = jnp.finfo(self._kernel_matrix.dtype).tiny
      cost = -jnp.log(self._kernel_matrix + eps)
      cost *= self.inv_scale_cost
      return cost if self._epsilon_init is None else self.epsilon * cost
    return self._cost_matrix * self.inv_scale_cost

  @property
  def median_cost_matrix(self) -> float:
    """Median of the :attr:`cost_matrix`."""
    return jnp.median(self.cost_matrix)

  @property
  def mean_cost_matrix(self) -> float:
    """Mean of the :attr:`cost_matrix`."""
    n, m = self.shape
    tmp = self.apply_cost(jnp.full((n,), fill_value=1.0 / n))
    return jnp.sum((1.0 / m) * tmp)

  @property
  def std_cost_matrix(self) -> float:
    r"""Standard deviation of all values stored in :attr:`cost_matrix`.

    Uses the :meth:`apply_square_cost` to remain
    applicable to low-rank matrices, through the formula:

    .. math::
        \sigma^2=\frac{1}{nm}\left(\sum_{ij} C_{ij}^2 -
        (\sum_{ij}C_ij)^2\right).

    to output :math:`\sigma`.
    """
    n, m = self.shape
    tmp = self.apply_square_cost(jnp.full((n,), fill_value=1.0 / n))
    tmp = jnp.sum((1.0 / m) * tmp) - (self.mean_cost_matrix ** 2)
    return jnp.sqrt(jax.nn.relu(tmp))

  @property
  def kernel_matrix(self) -> jnp.ndarray:
    """Kernel matrix.

    Either provided by user or recomputed from :attr:`cost_matrix`.
    """
    if self._kernel_matrix is None:
      return jnp.exp(-self._cost_matrix * self.inv_scale_cost / self.epsilon)
    return self._kernel_matrix ** self.inv_scale_cost

  @property
  def epsilon_scheduler(self) -> eps_scheduler.Epsilon:
    """Epsilon scheduler."""
    if isinstance(self._epsilon_init, eps_scheduler.Epsilon):
      return self._epsilon_init
    # no relative epsilon
    if self._relative_epsilon is None:
      if self._epsilon_init is not None:
        return eps_scheduler.Epsilon(self._epsilon_init)
      multiplier = eps_scheduler.DEFAULT_EPSILON_SCALE
      scale = jax.lax.stop_gradient(self.std_cost_matrix)
      return eps_scheduler.Epsilon(target=multiplier * scale)

    if self._relative_epsilon == "std":
      scale = jax.lax.stop_gradient(self.std_cost_matrix)
    elif self._relative_epsilon == "mean":
      scale = jax.lax.stop_gradient(self.mean_cost_matrix)
    else:
      raise ValueError(f"Invalid relative epsilon: {self._relative_epsilon}.")

    multiplier = (
        eps_scheduler.DEFAULT_EPSILON_SCALE
        if self._epsilon_init is None else self._epsilon_init
    )
    return eps_scheduler.Epsilon(target=multiplier * scale)

  @property
  def epsilon(self) -> float:
    """Epsilon regularization value."""
    return self.epsilon_scheduler.target

  @property
  def shape(self) -> Tuple[int, int]:
    """Shape of the geometry."""
    mat = (
        self._kernel_matrix if self._cost_matrix is None else self._cost_matrix
    )
    if mat is not None:
      return mat.shape
    return 0, 0

  @property
  def can_LRC(self) -> bool:
    """Check quickly if casting geometry as LRC makes sense.

    This check is only carried out using basic considerations from the geometry,
    not using a rigorous check involving, e.g., SVD.
    """
    return False

  @property
  def is_squared_euclidean(self) -> bool:
    """Whether cost is computed by taking squared Euclidean distance."""
    return False

  @property
  def is_online(self) -> bool:
    """Whether geometry cost/kernel should be recomputed on the fly."""
    return False

  @property
  def is_symmetric(self) -> bool:
    """Whether geometry cost/kernel is a symmetric matrix."""
    mat = self.kernel_matrix if self.cost_matrix is None else self.cost_matrix
    return self.is_square and jnp.all(mat == mat.T)

  @property
  def is_square(self) -> bool:
    """Whether geometry cost/kernel is a square matrix."""
    n, m = self.shape
    return (n == m)

  @property
  def inv_scale_cost(self) -> jnp.ndarray:
    """Compute and return inverse of scaling factor for cost matrix."""
    if self._scale_cost == "max_cost":
      return 1.0 / jnp.max(self._cost_matrix)
    if self._scale_cost == "mean":
      return 1.0 / jnp.mean(self._cost_matrix)
    if self._scale_cost == "median":
      return 1.0 / jnp.median(self._cost_matrix)
    if utils.is_scalar(self._scale_cost):
      return 1.0 / self._scale_cost
    raise ValueError(f"Scaling {self._scale_cost} not implemented.")

  @property
  def diag_cost(self) -> jnp.ndarray:
    """Diagonal of the cost matrix."""
    assert self.is_square, "Cost matrix must be square to compute diagonal."
    return jnp.diag(self.cost_matrix)

  def set_scale_cost(self, scale_cost: Union[float, str]) -> "Geometry":
    """Modify how to rescale of the :attr:`cost_matrix`."""
    # case when `geom` doesn't have `scale_cost` or doesn't need to be modified
    # `False` retains the original scale
    if scale_cost == self._scale_cost:
      return self
    children, aux_data = self.tree_flatten()
    aux_data["scale_cost"] = scale_cost
    return type(self).tree_unflatten(aux_data, children)

  def copy_epsilon(self, other: "Geometry") -> "Geometry":
    """Copy the epsilon parameters from another geometry."""
    children, aux_data = self.tree_flatten()
    new_geom = type(self).tree_unflatten(aux_data, children)
    new_geom._epsilon_init = other.epsilon_scheduler
    new_geom._relative_epsilon = other._relative_epsilon  # has no effect
    return new_geom

  # The functions below are at the core of Sinkhorn iterations, they
  # are implemented here in their default form, either in lse (using directly
  # cost matrices in stabilized form) or kernel mode (using kernel matrices).

  def apply_lse_kernel(
      self,
      f: jnp.ndarray,
      g: jnp.ndarray,
      eps: float,
      vec: jnp.ndarray = None,
      axis: int = 0
  ) -> Tuple[jnp.ndarray, jnp.ndarray]:
    r"""Apply :attr:`kernel_matrix` in log domain.

    This function applies the ground geometry's kernel in log domain, using
    a stabilized formulation. At a high level, this iteration performs either:

    - output = eps * log (K (exp(g / eps) * vec))  (1)
    - output = eps * log (K'(exp(f / eps) * vec))  (2)

    K is implicitly exp(-cost_matrix/eps).

    To carry this out in a stabilized way, we take advantage of the fact that
    the entries of the matrix ``f[:,*] + g[*,:] - C`` are all negative, and
    therefore their exponential never overflows, to add (and subtract after)
    f and g in iterations 1 & 2 respectively.

    Args:
      f: jnp.ndarray [num_a,] , potential of size num_rows of cost_matrix
      g: jnp.ndarray [num_b,] , potential of size num_cols of cost_matrix
      eps: float, regularization strength
      vec: jnp.ndarray [num_a or num_b,] , when not None, this has the effect of
        doing log-Kernel computations with an addition elementwise
        multiplication of exp(g / eps) by a vector. This is carried out by
        adding weights to the log-sum-exp function, and needs to handle signs
        separately.
      axis: summing over axis 0 when doing (2), or over axis 1 when doing (1)

    Returns:
      A jnp.ndarray corresponding to output above, depending on axis.
    """
    w_res, w_sgn = self._softmax(f, g, eps, vec, axis)
    remove = f if axis == 1 else g
    return w_res - jnp.where(jnp.isfinite(remove), remove, 0), w_sgn

  def apply_kernel(
      self,
      vec: jnp.ndarray,
      eps: Optional[float] = None,
      axis: int = 0,
  ) -> jnp.ndarray:
    """Apply :attr:`kernel_matrix` on positive scaling vector.

    Args:
      vec: jnp.ndarray [num_a or num_b] , scaling of size num_rows or
        num_cols of kernel_matrix
      eps: passed for consistency, not used yet.
      axis: standard kernel product if axis is 1, transpose if 0.

    Returns:
      a jnp.ndarray corresponding to output above, depending on axis.
    """
    if eps is None:
      kernel = self.kernel_matrix
    else:
      kernel = self.kernel_matrix ** (self.epsilon / eps)
    kernel = kernel if axis == 1 else kernel.T

    return jnp.dot(kernel, vec)

  def marginal_from_potentials(
      self,
      f: jnp.ndarray,
      g: jnp.ndarray,
      axis: int = 0,
  ) -> jnp.ndarray:
    """Output marginal of transportation matrix from potentials.

    This applies first lse kernel in the standard way, removes the
    correction used to stabilize computations, and lifts this with an exp to
    recover either of the marginals corresponding to the transport map induced
    by potentials.

    Args:
      f: jnp.ndarray [num_a,] , potential of size num_rows of cost_matrix
      g: jnp.ndarray [num_b,] , potential of size num_cols of cost_matrix
      axis: axis along which to integrate, returns marginal on other axis.

    Returns:
      a vector of marginals of the transport matrix.
    """
    h = (f if axis == 1 else g)
    z = self.apply_lse_kernel(f, g, self.epsilon, axis=axis)[0]
    return jnp.exp((z + h) / self.epsilon)

  def marginal_from_scalings(
      self,
      u: jnp.ndarray,
      v: jnp.ndarray,
      axis: int = 0,
  ) -> jnp.ndarray:
    """Output marginal of transportation matrix from scalings."""
    u, v = (v, u) if axis == 0 else (u, v)
    return u * self.apply_kernel(v, eps=self.epsilon, axis=axis)

  def transport_from_potentials(
      self, f: jnp.ndarray, g: jnp.ndarray
  ) -> jnp.ndarray:
    """Output transport matrix from potentials."""
    return jnp.exp(self._center(f, g) / self.epsilon)

  def transport_from_scalings(
      self, u: jnp.ndarray, v: jnp.ndarray
  ) -> jnp.ndarray:
    """Output transport matrix from pair of scalings."""
    return self.kernel_matrix * u[:, jnp.newaxis] * v[jnp.newaxis, :]

  # Functions that are not supposed to be changed by inherited classes.
  # These are the point of entry for Sinkhorn's algorithm to use a geometry.

  def update_potential(
      self,
      f: jnp.ndarray,
      g: jnp.ndarray,
      log_marginal: jnp.ndarray,
      iteration: Optional[int] = None,
      axis: int = 0,
  ) -> jnp.ndarray:
    """Carry out one Sinkhorn update for potentials, i.e. in log space.

    Args:
      f: jnp.ndarray [num_a,] , potential of size num_rows of cost_matrix
      g: jnp.ndarray [num_b,] , potential of size num_cols of cost_matrix
      log_marginal: targeted marginal
      iteration: used to compute epsilon from schedule, if provided.
      axis: axis along which the update should be carried out.

    Returns:
      new potential value, g if axis=0, f if axis is 1.
    """
    eps = self.epsilon_scheduler(iteration)
    app_lse = self.apply_lse_kernel(f, g, eps, axis=axis)[0]
    return eps * log_marginal - jnp.where(jnp.isfinite(app_lse), app_lse, 0)

  def update_scaling(
      self,
      scaling: jnp.ndarray,
      marginal: jnp.ndarray,
      iteration: Optional[int] = None,
      axis: int = 0,
  ) -> jnp.ndarray:
    """Carry out one Sinkhorn update for scalings, using kernel directly.

    Args:
      scaling: jnp.ndarray of num_a or num_b positive values.
      marginal: targeted marginal
      iteration: used to compute epsilon from schedule, if provided.
      axis: axis along which the update should be carried out.

    Returns:
      new scaling vector, of size num_b if axis=0, num_a if axis is 1.
    """
    eps = self.epsilon_scheduler(iteration)
    app_kernel = self.apply_kernel(scaling, eps, axis=axis)
    return marginal / jnp.where(app_kernel > 0, app_kernel, 1.0)

  # Helper functions
  def _center(self, f: jnp.ndarray, g: jnp.ndarray) -> jnp.ndarray:
    return f[:, jnp.newaxis] + g[jnp.newaxis, :] - self.cost_matrix

  def _softmax(
      self, f: jnp.ndarray, g: jnp.ndarray, eps: float,
      vec: Optional[jnp.ndarray], axis: int
  ) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Apply softmax row or column wise, weighted by vec."""
    if vec is not None:
      if axis == 0:
        vec = vec.reshape((-1, 1))
      lse_output = mu.logsumexp(
          self._center(f, g) / eps, b=vec, axis=axis, return_sign=True
      )
      return eps * lse_output[0], lse_output[1]

    lse_output = mu.logsumexp(
        self._center(f, g) / eps, axis=axis, return_sign=False
    )
    return eps * lse_output, jnp.array([1.0])

  @functools.partial(jax.vmap, in_axes=[None, None, None, 0, None])
  def _apply_transport_from_potentials(
      self, f: jnp.ndarray, g: jnp.ndarray, vec: jnp.ndarray, axis: int
  ) -> jnp.ndarray:
    """Apply lse_kernel to arbitrary vector while keeping track of signs."""
    lse_res, lse_sgn = self.apply_lse_kernel(
        f, g, self.epsilon, vec=vec, axis=axis
    )
    lse_res += f if axis == 1 else g
    return lse_sgn * jnp.exp(lse_res / self.epsilon)

  # wrapper to allow default option for axis.
  def apply_transport_from_potentials(
      self,
      f: jnp.ndarray,
      g: jnp.ndarray,
      vec: jnp.ndarray,
      axis: int = 0
  ) -> jnp.ndarray:
    """Apply transport matrix computed from potentials to a (batched) vec.

    This approach does not instantiate the transport matrix itself, but uses
    instead potentials to apply the transport using apply_lse_kernel, therefore
    guaranteeing stability and lower memory footprint.

    Computations are done in log space, and take advantage of the
    (b=..., return_sign=True) optional parameters of logsumexp.

    Args:
      f: jnp.ndarray [num_a,] , potential of size num_rows of cost_matrix
      g: jnp.ndarray [num_b,] , potential of size num_cols of cost_matrix
      vec: jnp.ndarray [batch, num_a or num_b], vector that will be multiplied
        by transport matrix corresponding to potentials f, g, and geom.
      axis: axis to differentiate left (0) or right (1) multiply.

    Returns:
      ndarray of the size of vec.
    """
    if vec.ndim == 1:
      return self._apply_transport_from_potentials(
          f, g, vec[jnp.newaxis, :], axis
      )[0, :]
    return self._apply_transport_from_potentials(f, g, vec, axis)

  @functools.partial(jax.vmap, in_axes=[None, None, None, 0, None])
  def _apply_transport_from_scalings(
      self, u: jnp.ndarray, v: jnp.ndarray, vec: jnp.ndarray, axis: int
  ):
    u, v = (u, v * vec) if axis == 1 else (v, u * vec)
    return u * self.apply_kernel(v, eps=self.epsilon, axis=axis)

  # wrapper to allow default option for axis
  def apply_transport_from_scalings(
      self,
      u: jnp.ndarray,
      v: jnp.ndarray,
      vec: jnp.ndarray,
      axis: int = 0
  ) -> jnp.ndarray:
    """Apply transport matrix computed from scalings to a (batched) vec.

    This approach does not instantiate the transport matrix itself, but
    relies instead on the apply_kernel function.

    Args:
      u: jnp.ndarray [num_a,] , scaling of size num_rows of cost_matrix
      v: jnp.ndarray [num_b,] , scaling of size num_cols of cost_matrix
      vec: jnp.ndarray [batch, num_a or num_b], vector that will be multiplied
        by transport matrix corresponding to scalings u, v, and geom.
      axis: axis to differentiate left (0) or right (1) multiply.

    Returns:
      ndarray of the size of vec.
    """
    if vec.ndim == 1:
      return self._apply_transport_from_scalings(
          u, v, vec[jnp.newaxis, :], axis
      )[0, :]
    return self._apply_transport_from_scalings(u, v, vec, axis)

  def potential_from_scaling(self, scaling: jnp.ndarray) -> jnp.ndarray:
    """Compute dual potential vector from scaling vector.

    Args:
      scaling: vector.

    Returns:
      a vector of the same size.
    """
    return self.epsilon * jnp.log(scaling)

  def scaling_from_potential(self, potential: jnp.ndarray) -> jnp.ndarray:
    """Compute scaling vector from dual potential.

    Args:
      potential: vector.

    Returns:
      a vector of the same size.
    """
    finite = jnp.isfinite(potential)
    return jnp.where(
        finite, jnp.exp(jnp.where(finite, potential / self.epsilon, 0.0)), 0.0
    )

  def apply_square_cost(self, arr: jnp.ndarray, axis: int = 0) -> jnp.ndarray:
    """Apply elementwise-square of cost matrix to array (vector or matrix).

    This function applies the ground geometry's cost matrix, to perform either
    output = C arr (if axis=1)
    output = C' arr (if axis=0)
    where C is [num_a, num_b], when the cost matrix itself is computed as a
    squared-Euclidean distance between vectors, and therefore admits an
    explicit low-rank factorization.

    Args:
      arr: array.
      axis: axis of the array on which the cost matrix should be applied.

    Returns:
      An array, [num_b, p] if axis=0 or [num_a, p] if axis=1.
    """
    return self.apply_cost(arr, axis=axis, fn=lambda x: x ** 2)

  def apply_cost(
      self,
      arr: jnp.ndarray,
      axis: int = 0,
      fn: Optional[Callable[[jnp.ndarray], jnp.ndarray]] = None,
      is_linear: bool = False,
  ) -> jnp.ndarray:
    """Apply :attr:`cost_matrix` to array (vector or matrix).

    This function applies the ground geometry's cost matrix, to perform either
    output = C arr (if axis=1)
    output = C' arr (if axis=0)
    where C is [num_a, num_b]

    Args:
      arr: jnp.ndarray [num_a or num_b, p], vector that will be multiplied by
        the cost matrix.
      axis: standard cost matrix if axis=1, transpose if 0
      fn: function to apply to cost matrix element-wise before the dot product
      is_linear: Whether ``fn`` is linear.

    Returns:
      An array, [num_b, p] if axis=0 or [num_a, p] if axis=1
    """
    if arr.ndim == 1:
      return self._apply_cost_to_vec(arr, axis=axis, fn=fn, is_linear=is_linear)
    app = functools.partial(
        self._apply_cost_to_vec, axis=axis, fn=fn, is_linear=is_linear
    )
    return jax.vmap(app, in_axes=1, out_axes=1)(arr)

  def _apply_cost_to_vec(
      self,
      vec: jnp.ndarray,
      axis: int = 0,
      fn: Optional[Callable[[jnp.ndarray], jnp.ndarray]] = None,
      is_linear: bool = False,
  ) -> jnp.ndarray:
    """Apply ``[num_a, num_b]`` fn(cost) (or transpose) to vector.

    Args:
      vec: jnp.ndarray [num_a,] ([num_b,] if axis=1) vector
      axis: axis on which the reduction is done.
      fn: function optionally applied to cost matrix element-wise, before the
        doc product
      is_linear: Whether ``fn`` is linear.

    Returns:
      A jnp.ndarray corresponding to cost x vector
    """
    del is_linear
    matrix = self.cost_matrix.T if axis == 0 else self.cost_matrix
    if fn is not None:
      matrix = fn(matrix)
    return jnp.dot(matrix, vec)

  @classmethod
  def prepare_divergences(
      cls,
      *args: Any,
      static_b: bool = False,
      **kwargs: Any
  ) -> Tuple["Geometry", ...]:
    """Instantiate 2 (or 3) geometries to compute a Sinkhorn divergence."""
    size = 2 if static_b else 3
    nones = [None, None, None]
    cost_matrices = kwargs.pop("cost_matrix", args)
    kernel_matrices = kwargs.pop("kernel_matrix", nones)
    cost_matrices = cost_matrices if cost_matrices is not None else nones
    return tuple(
        cls(cost_matrix=arg1, kernel_matrix=arg2, **kwargs)
        for arg1, arg2, _ in zip(cost_matrices, kernel_matrices, range(size))
    )

  def to_LRCGeometry(
      self,
      rank: int = 0,
      tol: float = 1e-2,
      rng: Optional[jax.Array] = None,
      scale: float = 1.0
  ) -> "low_rank.LRCGeometry":
    r"""Factorize the cost matrix using either SVD (full) or :cite:`indyk:19`.

    When `rank=min(n,m)` or `0` (by default), use :func:`jax.numpy.linalg.svd`.

    For other values, use the routine in sublinear time :cite:`indyk:19`.
    Uses the implementation of :cite:`scetbon:21`, algorithm 4.

    It holds that with probability *0.99*,
    :math:`||A - UV||_F^2 \leq || A - A_k ||_F^2 + tol \cdot ||A||_F^2`,
    where :math:`A` is ``n x m`` cost matrix, :math:`UV` the factorization
    computed in sublinear time and :math:`A_k` the best rank-k approximation.

    Args:
      rank: Target rank of the :attr:`cost_matrix`.
      tol: Tolerance of the error. The total number of sampled points is
        :math:`min(n, m,\frac{rank}{tol})`.
      rng: The PRNG key to use for initializing the model.
      scale: Value used to rescale the factors of the low-rank geometry.
        Useful when this geometry is used in the linear term of fused GW.

    Returns:
      Low-rank geometry.
    """
    from ott.geometry import low_rank
    assert rank >= 0, f"Rank must be non-negative, got {rank}."
    n, m = self.shape

    if rank == 0 or rank >= min(n, m):
      # TODO(marcocuturi): add hermitian=self.is_symmetric, currently bugging.
      u, s, vh = jnp.linalg.svd(
          self.cost_matrix,
          full_matrices=False,
          compute_uv=True,
      )

      cost_1 = u
      cost_2 = (s[:, None] * vh).T
    else:
      rng = utils.default_prng_key(rng)
      rng1, rng2, rng3, rng4, rng5 = jax.random.split(rng, 5)
      n_subset = min(int(rank / tol), n, m)

      i_star = jax.random.randint(rng1, shape=(), minval=0, maxval=n)
      j_star = jax.random.randint(rng2, shape=(), minval=0, maxval=m)

      ci_star = self.subset(row_ixs=i_star).cost_matrix.ravel() ** 2  # (m,)
      cj_star = self.subset(col_ixs=j_star).cost_matrix.ravel() ** 2  # (n,)

      p_row = cj_star + ci_star[j_star] + jnp.mean(ci_star)  # (n,)
      p_row /= jnp.sum(p_row)
      row_ixs = jax.random.choice(rng3, n, shape=(n_subset,), p=p_row)
      # (n_subset, m)
      s = self.subset(row_ixs=row_ixs).cost_matrix
      s /= jnp.sqrt(n_subset * p_row[row_ixs][:, None])

      p_col = jnp.sum(s ** 2, axis=0)  # (m,)
      p_col /= jnp.sum(p_col)
      # (n_subset,)
      col_ixs = jax.random.choice(rng4, m, shape=(n_subset,), p=p_col)
      # (n_subset, n_subset)
      w = s[:, col_ixs] / jnp.sqrt(n_subset * p_col[col_ixs][None, :])

      U, _, V = jsp.linalg.svd(w)
      U = U[:, :rank]  # (n_subset, rank)
      U = (s.T @ U) / jnp.linalg.norm(w.T @ U, axis=0)  # (m, rank)

      _, d, v = jnp.linalg.svd(U.T @ U)  # (k,), (k, k)
      v = v.T / jnp.sqrt(d)[None, :]

      inv_scale = (1.0 / jnp.sqrt(n_subset))
      col_ixs = jax.random.choice(rng5, m, shape=(n_subset,))  # (n_subset,)

      # (n, n_subset)
      A_trans = self.subset(col_ixs=col_ixs).cost_matrix * inv_scale
      B = (U[col_ixs, :] @ v * inv_scale)  # (n_subset, k)
      M = jnp.linalg.inv(B.T @ B)  # (k, k)
      V = jnp.linalg.multi_dot([A_trans, B, M.T, v.T])  # (n, k)
      cost_1 = V
      cost_2 = U

    return low_rank.LRCGeometry(
        cost_1=cost_1,
        cost_2=cost_2,
        epsilon=self._epsilon_init,
        relative_epsilon=self._relative_epsilon,
        scale_cost=self._scale_cost,
        scale_factor=scale,
    )

  def subset(
      self,
      row_ixs: Optional[jnp.ndarray] = None,
      col_ixs: Optional[jnp.ndarray] = None
  ) -> "Geometry":
    """Subset rows or columns of a geometry.

    Args:
      row_ixs: Row indices. If :obj:`None`, use all rows.
      col_ixs: Column indices. If :obj:`None`, use all columns.

    Returns:
      The subsetted geometry.
    """
    (cost, kernel, *rest), aux_data = self.tree_flatten()
    row_ixs = row_ixs if row_ixs is None else jnp.atleast_1d(row_ixs)
    col_ixs = col_ixs if col_ixs is None else jnp.atleast_1d(col_ixs)
    if cost is not None:
      cost = cost if row_ixs is None else cost[row_ixs]
      cost = cost if col_ixs is None else cost[:, col_ixs]
    if kernel is not None:
      kernel = kernel if row_ixs is None else kernel[row_ixs]
      kernel = kernel if col_ixs is None else kernel[:, col_ixs]
    return type(self).tree_unflatten(aux_data, (cost, kernel, *rest))

  @property
  def dtype(self) -> jnp.dtype:
    """The data type."""
    if self._cost_matrix is not None:
      return self._cost_matrix.dtype
    return self._kernel_matrix.dtype

  def tree_flatten(self):  # noqa: D102
    return (
        self._cost_matrix,
        self._kernel_matrix,
        self._epsilon_init,
    ), {
        "scale_cost": self._scale_cost,
        "relative_epsilon": self._relative_epsilon,
    }

  @classmethod
  def tree_unflatten(cls, aux_data, children):  # noqa: D102
    cost, kernel, epsilon = children
    return cls(cost, kernel_matrix=kernel, epsilon=epsilon, **aux_data)



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from . import fixed_point_loop, matrix_square_root, unbalanced_functions, utils
from ._lbfgs import lbfgs
from ._legendre import legendre
from ._velocity_from_brenier_potential import velocity_from_brenier_potential

__all__ = [
    "fixed_point_loop",
    "matrix_square_root",
    "unbalanced_functions",
    "utils",
    "lbfgs",
    "legendre",
    "velocity_from_brenier_potential",
]
